/*
 gpuSEM.c
 
 Main functions for running simulation.

 Author: Scott Christley <schristley@mac.com>
 
 Copyright (C) 2010 Scott Christley
 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met: 1. Redistributions of source code must retain the above
 copyright notice, this list of conditions and the following
 disclaimer.  2. Redistributions in binary form must reproduce the
 above copyright notice, this list of conditions and the following
 disclaimer in the documentation and/or other materials provided with
 the distribution.  3. The name of the author may not be used to
 endorse or promote products derived from this software without
 specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 */


#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "GPUDefines.h"

#define RAND_NUM ((double)rand() / (double)RAND_MAX )

// Generate normally distributed random number
double normal_dist_rand(double mean, double sd)
{
  double fac, radius, v1, v2;
  double rd1Value, rd2Value;

  do {
    rd1Value = RAND_NUM;
    rd2Value = RAND_NUM;
    v1 = (2.0 * rd1Value) - 1.0;
    v2 = (2.0 * rd2Value) - 1.0;
    radius = v1*v1 + v2*v2; 
  } while (radius >= 1.0);
  fac = sqrt (-2.0 * log (radius) / radius);
  return v2 * fac * sd + mean;    // use fixed params
}


// The GPU functions
extern void *sem_allocGPUKernel(void *model, int maxCells, int maxElements, float dt, int numOfSpecies, float *hostParameters);
extern void sem_initGPUKernel(void *model, void *g, int numOfCells, int *numOfElements, void *hostX, void *hostY, void *hostZ, void *hostType, float *speciesData);
extern void sem_invokeGPUKernel(void *model, void *g, void *hostX, void *hostY, void *hostZ,
				void *, void *, void *, void *, void *, int timeSteps);
extern void sem_releaseGPUKernel(void *model, void *g);

float hostParameters[100];

// spatial domain
// For 128 and 250 cells
#define BOUNDARY_X 5.9
#define BOUNDARY_Y 5.9
// For 500 cells
//#define BOUNDARY_X 9.9
//#define BOUNDARY_Y 9.9
// For 1000 cells
//#define BOUNDARY_X 14.9
//#define BOUNDARY_Y 14.9
// For 5000 cells
//#define BOUNDARY_X 29.9
//#define BOUNDARY_Y 29.9

#define BOUNDARY_Z 10.0
#define Z_RANGE 0.5
#define Z_OFFSET 0.5

#define INTER_DIST 0.6

// species index
#define NOTCH_species 0
#define DELTA_species 1
#define BOUND_species 2
#define BMA_species 3
#define OVOL1_species 4
#define OVOL2_species 5
#define CMYC_species 6


void runGPUSimulation()
{
  int i, j;
  int p, l, k;

  int maxCells = 1000;
  int numOfCells = 100;
  int maxElements = 100;
  int initialElements = 20;
  int numOfSpecies = 8;

  // initialize cells
  int *numOfElements = (int *)malloc(maxCells * sizeof(int));
  void *elementType = malloc(maxCells * maxElements * sizeof(int));
  int (*etGrid)[maxElements][maxCells] = elementType;

  void *X = malloc(maxCells * maxElements * sizeof(float));
  void *Y = malloc(maxCells * maxElements * sizeof(float));
  void *Z = malloc(maxCells * maxElements * sizeof(float));
  float (*xGrid)[maxElements][maxCells] = X;
  float (*yGrid)[maxElements][maxCells] = Y;
  float (*zGrid)[maxElements][maxCells] = Z;

  float *cX = (float *)malloc(maxCells * sizeof(float));
  float *cY = (float *)malloc(maxCells * sizeof(float));
  float *cZ = (float *)malloc(maxCells * sizeof(float));

  void *numNeighbors = malloc(maxCells * maxElements * sizeof(int));
  int (*nnGrid)[maxElements][maxCells] = numNeighbors;

  void *speciesData = (float *)malloc(maxCells * numOfSpecies * sizeof(float));
  float (*sData)[numOfSpecies][maxCells] = speciesData;

  // initial values
  for (j = 0; j < numOfCells; ++j) {
    (*sData)[NOTCH_species][j] = normal_dist_rand(1, 1);
    if ((*sData)[NOTCH_species][j] < 0) (*sData)[NOTCH_species][j] = 0;
    (*sData)[DELTA_species][j] = normal_dist_rand(1, 1);
    if ((*sData)[DELTA_species][j] < 0) (*sData)[DELTA_species][j] = 0;
    (*sData)[BOUND_species][j] = 0;
    (*sData)[OVOL1_species][j] = 0;
    (*sData)[OVOL2_species][j] = 0;
    (*sData)[CMYC_species][j] = 0;

    cX[j] = 0.0;
    cY[j] = 0.0;
    cZ[j] = 0.0;
    for (i = 0; i < maxElements; ++i)
      (*etGrid)[i][j] = 0;
  }

  // initial random positions for elements
  for (j = 0; j < numOfCells; ++j) {
    numOfElements[j] = initialElements;

    // place initial element
    (*xGrid)[0][j] = BOUNDARY_X * RAND_NUM;
    if ((*xGrid)[0][j] < 0.0) (*xGrid)[0][j] = 0.0;
    if ((*xGrid)[0][j] > BOUNDARY_X) (*xGrid)[0][j] = BOUNDARY_X;

    (*yGrid)[0][j] = BOUNDARY_Y * RAND_NUM;
    if ((*yGrid)[0][j] < 0.0) (*yGrid)[0][j] = 0.0;
    if ((*yGrid)[0][j] > BOUNDARY_Y) (*yGrid)[0][j] = BOUNDARY_Y;

    (*zGrid)[0][j] = Z_RANGE * RAND_NUM + Z_OFFSET;
    if ((*zGrid)[0][j] < 0.0) (*zGrid)[0][j] = 0.0;
    if ((*zGrid)[0][j] > BOUNDARY_Z) (*zGrid)[0][j] = BOUNDARY_Z;

    // place remaining elements near initial element
    for (i = 1; i < numOfElements[j]; ++i) {
      (*xGrid)[i][j] = (*xGrid)[0][j] + RAND_NUM * 0.1 - 0.05;
      if ((*xGrid)[i][j] < 0.0) (*xGrid)[i][j] = 0.0;
      if ((*xGrid)[i][j] > BOUNDARY_X) (*xGrid)[i][j] = BOUNDARY_X;

      (*yGrid)[i][j] = (*yGrid)[0][j] + RAND_NUM * 0.1 - 0.05;
      if ((*yGrid)[i][j] < 0.0) (*yGrid)[i][j] = 0.0;
      if ((*yGrid)[i][j] > BOUNDARY_Y) (*yGrid)[i][j] = BOUNDARY_Y;

      (*zGrid)[i][j] = (*zGrid)[0][j] + RAND_NUM * 0.1 - 0.05;
      if ((*zGrid)[i][j] < 0.0) (*zGrid)[i][j] = 0.0;
      if ((*zGrid)[i][j] > BOUNDARY_Z) (*zGrid)[i][j] = BOUNDARY_Z;
    }
  }

  // basement membrane adhesion
  // set element type
  for (j = 0; j < numOfCells; ++j)
    for (i = 0; i < 4; ++i)
      (*etGrid)[i][j] = 1;

  // intracellular gene network

  // notch-delta parameters
  // lateral inhibition
  hostParameters[0] = 0.0003;
  hostParameters[1] = 0.12;
  hostParameters[2] = 0.019;
  hostParameters[3] = 0.006;
  hostParameters[4] = 0.03;
  // Hill parameters, BOUND on NOTCH
  hostParameters[5] = 0.01;
  hostParameters[6] = 1.0;
  hostParameters[7] = 1.0;
  hostParameters[8] = -2.0;
  // Hill parameters, BOUND on DELTA
  hostParameters[9] = 0.01;
  hostParameters[10] = 1.0;
  hostParameters[11] = 0.5;
  hostParameters[12] = 2.0;

  // Hill parameters, BMA on OVOL1
  hostParameters[13] = 0.0;
  hostParameters[14] = 1.0;
  hostParameters[15] = 1.0;
  hostParameters[16] = 1.0;

  // Hill parameters, OVOL1 on OVOL2
  hostParameters[17] = 0.1;
  hostParameters[18] = 2.0;
  hostParameters[19] = 1.0;
  hostParameters[20] = 2.0;

  // Hill parameters, OVOL2 on OVOL1
  hostParameters[21] = 0.0;
  hostParameters[22] = 2.0;
  hostParameters[23] = 1.0;
  hostParameters[24] = 2.0;

  // Hill parameters, OVOL2 on NOTCH
  hostParameters[25] = 0.0;
  hostParameters[26] = 1.0;
  hostParameters[27] = 2.0;
  hostParameters[28] = 1.0;

  // Hill parameters, OVOL2 on CMYC
  hostParameters[29] = 0.1;
  hostParameters[30] = 1.0;
  hostParameters[31] = 1.0;
  hostParameters[32] = 1.0;

  // linear interactions
  hostParameters[33] = 1.0; // OVOL1 decay
  hostParameters[34] = 1.0; // OVOL2 decay
  hostParameters[35] = 1.0; // CMYC decay

  // load initial conditions
  FILE *initialData = fopen("initial.dat", "rb");
    fread(&(numOfCells), sizeof(int), 1, initialData);
    for (j = 0; j < numOfCells; ++j) {
      fread(&(numOfElements[j]), sizeof(int), 1, initialData);
      for (i = 0; i < numOfElements[j]; ++i) {
	fread(&((*xGrid)[i][j]), sizeof(float), 1, initialData);
	fread(&((*yGrid)[i][j]), sizeof(float), 1, initialData);
	fread(&((*zGrid)[i][j]), sizeof(float), 1, initialData);
      }
    }
  fclose(initialData);
  printf("numOfCells = %d\n", numOfCells);

  // execution settings
  float dt = 0.1;
  int simulateTime = 13000;
  int timeSteps = simulateTime / dt;
  int outputFrequency = 100;

  // initialize GPU
  void *gpuGrids = sem_allocGPUKernel(NULL, maxCells, maxElements, dt, numOfSpecies, hostParameters);
  sem_initGPUKernel(NULL, gpuGrids, numOfCells, numOfElements, X, Y, Z, elementType, speciesData);

  // output data files
  FILE *dataFiles[2];

  dataFiles[0] = fopen("cells.data", "wb+");
  dataFiles[1] = fopen("network.data", "wb+");

  fwrite(&(numOfCells), sizeof(int), 1, dataFiles[0]);
  for (j = 0; j < numOfCells; ++j) {
    fwrite(&(numOfElements[j]), sizeof(int), 1, dataFiles[0]);
    fwrite(&((*sData)[NOTCH_species][j]), sizeof(float), 1, dataFiles[1]);
    fwrite(&((*sData)[DELTA_species][j]), sizeof(float), 1, dataFiles[1]);
    fwrite(&((*sData)[BOUND_species][j]), sizeof(float), 1, dataFiles[1]);
    fwrite(&((*sData)[OVOL1_species][j]), sizeof(float), 1, dataFiles[1]);
    fwrite(&((*sData)[OVOL2_species][j]), sizeof(float), 1, dataFiles[1]);
    fwrite(&((*sData)[CMYC_species][j]), sizeof(float), 1, dataFiles[1]);
    for (i = 0; i < numOfElements[j]; ++i) {
      fwrite(&((*xGrid)[i][j]), sizeof(float), 1, dataFiles[0]);
      fwrite(&((*yGrid)[i][j]), sizeof(float), 1, dataFiles[0]);
      fwrite(&((*zGrid)[i][j]), sizeof(float), 1, dataFiles[0]);
      fwrite(&((*etGrid)[i][j]), sizeof(int), 1, dataFiles[0]);
    }
  }

  // run it
  int done = 0;
  int t = 0;
  while (!done) {

    // invoke GPU
#if 0
      for (j = 0; j < numOfCells; ++j) {
	printf("(%d) N/D/B/A/1/2/C = (%f, %f, %f, %f, %f, %f, %f)\n", j, (*sData)[NOTCH_species][j], (*sData)[DELTA_species][j],
	       (*sData)[BOUND_species][j], (*sData)[BMA_species][j], (*sData)[OVOL1_species][j], (*sData)[OVOL2_species][j],
	       (*sData)[CMYC_species][j]);
      }
#endif
    // movement and network
    sem_invokeGPUKernel(NULL, gpuGrids, X, Y, Z, speciesData, cX, cY, cZ, numNeighbors, outputFrequency);

    // are we done?
    t += outputFrequency;
    if (t >= timeSteps) done = 1;

    // output data grids
    printf("time = %d\n", t);
    fwrite(&(numOfCells), sizeof(int), 1, dataFiles[0]);
    for (j = 0; j < numOfCells; ++j) {
      fwrite(&(numOfElements[j]), sizeof(int), 1, dataFiles[0]);
      fwrite(&((*sData)[NOTCH_species][j]), sizeof(float), 1, dataFiles[1]);
      fwrite(&((*sData)[DELTA_species][j]), sizeof(float), 1, dataFiles[1]);
      fwrite(&((*sData)[BOUND_species][j]), sizeof(float), 1, dataFiles[1]);
      fwrite(&((*sData)[OVOL1_species][j]), sizeof(float), 1, dataFiles[1]);
      fwrite(&((*sData)[OVOL2_species][j]), sizeof(float), 1, dataFiles[1]);
      fwrite(&((*sData)[CMYC_species][j]), sizeof(float), 1, dataFiles[1]);
      for (i = 0; i < numOfElements[j]; ++i) {
	fwrite(&((*xGrid)[i][j]), sizeof(float), 1, dataFiles[0]);
	fwrite(&((*yGrid)[i][j]), sizeof(float), 1, dataFiles[0]);
	fwrite(&((*zGrid)[i][j]), sizeof(float), 1, dataFiles[0]);
	fwrite(&((*etGrid)[i][j]), sizeof(int), 1, dataFiles[0]);
      }
    }


    // add element
    if (t%2000 == 0) {
      for (j = 0; j < numOfCells; ++j) {
	if ((*sData)[DELTA_species][j] < 100.0) continue;

	// put new element at center
	(*etGrid)[numOfElements[j]][j] = 0;
	(*xGrid)[numOfElements[j]][j] = cX[j];
	(*yGrid)[numOfElements[j]][j] = cY[j];
	(*zGrid)[numOfElements[j]][j] = cZ[j];
	printf("add element (%d, %d), (%f %f %f) at (%f, %f, %f)\n", j, numOfElements[j], 
	       (*xGrid)[0][j], (*yGrid)[0][j], (*zGrid)[0][j], cX[j], cY[j], cZ[j]);

	++numOfElements[j];
      }
      sem_initGPUKernel(NULL, gpuGrids, numOfCells, numOfElements, X, Y, Z, elementType, speciesData);
    }



    // divide cell
    for (j = 0; j < numOfCells; ++j) {
      if (numOfElements[j] != 40) continue;

      printf("cell division (%d) ", j);
      for (i = 20; i < numOfElements[j]; ++i) {
	(*etGrid)[i-20][numOfCells] = 0;

	(*xGrid)[i-20][numOfCells] = (*xGrid)[i][j];
	(*yGrid)[i-20][numOfCells] = (*yGrid)[i][j];
	(*zGrid)[i-20][numOfCells] = (*zGrid)[i][j];
      }
      numOfElements[j] -= 20;
      numOfElements[numOfCells] = 20;

      // divide Notch,Delta,Bound in half
      printf("(%f, %f, %f) -> ", (*sData)[NOTCH_species][j], (*sData)[DELTA_species][j], (*sData)[BOUND_species][j]);
      (*sData)[NOTCH_species][numOfCells] = 0.5 * (*sData)[NOTCH_species][j];
      (*sData)[NOTCH_species][j] = (*sData)[NOTCH_species][numOfCells];
      (*sData)[DELTA_species][numOfCells] = 0.5 * (*sData)[DELTA_species][j];
      (*sData)[DELTA_species][j] = (*sData)[DELTA_species][numOfCells];
      (*sData)[BOUND_species][numOfCells] = 0.5 * (*sData)[BOUND_species][j];
      (*sData)[BOUND_species][j] = (*sData)[BOUND_species][numOfCells];

      (*sData)[OVOL1_species][numOfCells] = 0.5 * (*sData)[OVOL1_species][j];
      (*sData)[OVOL1_species][j] = (*sData)[OVOL1_species][numOfCells];
      (*sData)[OVOL2_species][numOfCells] = 0.5 * (*sData)[OVOL2_species][j];
      (*sData)[OVOL2_species][j] = (*sData)[OVOL2_species][numOfCells];
      (*sData)[CMYC_species][numOfCells] = 0.5 * (*sData)[CMYC_species][j];
      (*sData)[CMYC_species][j] = (*sData)[CMYC_species][numOfCells];

      printf("(%d) N/D/B/A/1/2/C (%d) = (%f, %f, %f, %f, %f, %f, %f)\n", numOfCells, j, (*sData)[NOTCH_species][j], (*sData)[DELTA_species][j],
	     (*sData)[BOUND_species][j], (*sData)[BMA_species][j], (*sData)[OVOL1_species][j], (*sData)[OVOL2_species][j],
	     (*sData)[CMYC_species][j]);
      ++numOfCells;

      //printf("cells: %d\n", numOfCells);
      sem_initGPUKernel(NULL, gpuGrids, numOfCells, numOfElements, X, Y, Z, elementType, speciesData);
    }


  }

  // release GPU
  sem_releaseGPUKernel(NULL, gpuGrids);

  // write out final data
  fclose(dataFiles[0]);
  fclose(dataFiles[1]);
}

int
main( int argc, char** argv) 
{
  runGPUSimulation();
}
