/*
 sem_kernel.cu
 
 GPU kernel functions.

 Author: Scott Christley <schristley@mac.com>
 
 Copyright (C) 2010 Scott Christley
 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met: 1. Redistributions of source code must retain the above
 copyright notice, this list of conditions and the following
 disclaimer.  2. Redistributions in binary form must reproduce the
 above copyright notice, this list of conditions and the following
 disclaimer in the documentation and/or other materials provided with
 the distribution.  3. The name of the author may not be used to
 endorse or promote products derived from this software without
 specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 */


// parameters

#define INTRA_U0 0.3
//#define INTRA_U0 0.25
#define INTRA_U1 0.1
#define INTRA_ETA0 0.12
#define INTRA_ETA1 0.36

#define INTER_DIST 0.6
#define INTER_U0 0.3
//#define INTER_U0 0.25
#define INTER_U1 0.05
#define INTER_ETA0 0.12
#define INTER_ETA1 0.24

//#define NEIGHBOR_DIST 0.6
#define NEIGHBOR_DIST 0.8

#define KA skin_parameters[0]
#define KD skin_parameters[1]
#define KI skin_parameters[2]
#define DA skin_parameters[3]
#define DF skin_parameters[4]

// Hill parameters, NOTCH
#define NOTCH_pmin_BOUND skin_parameters[5]
#define NOTCH_pmax_BOUND skin_parameters[6]
#define NOTCH_c_BOUND skin_parameters[7]
#define NOTCH_h_BOUND skin_parameters[8]

// Hill parameters, DELTA
#define DELTA_pmin_BOUND skin_parameters[9]
#define DELTA_pmax_BOUND skin_parameters[10]
#define DELTA_c_BOUND skin_parameters[11]
#define DELTA_h_BOUND skin_parameters[12]

// Hill parameters, BMA on OVOL1
#define OVOL1_pmin_BMA skin_parameters[13]
#define OVOL1_pmax_BMA skin_parameters[14]
#define OVOL1_c_BMA skin_parameters[15]
#define OVOL1_h_BMA skin_parameters[16]

// Hill parameters, OVOL1 on OVOL2
#define OVOL2_pmin_OVOL1 skin_parameters[17]
#define OVOL2_pmax_OVOL1 skin_parameters[18]
#define OVOL2_c_OVOL1 skin_parameters[19]
#define OVOL2_h_OVOL1 skin_parameters[20]

// Hill parameters, OVOL2 on OVOL1
#define OVOL1_pmin_OVOL2 skin_parameters[21]
#define OVOL1_pmax_OVOL2 skin_parameters[22]
#define OVOL1_c_OVOL2 skin_parameters[23]
#define OVOL1_h_OVOL2 skin_parameters[24]

// Hill parameters, OVOL2 on NOTCH
#define NOTCH_pmin_OVOL2 skin_parameters[25]
#define NOTCH_pmax_OVOL2 skin_parameters[26]
#define NOTCH_c_OVOL2 skin_parameters[27]
#define NOTCH_h_OVOL2 skin_parameters[28]

// Hill parameters, OVOL2 on CMYC
#define CMYC_pmin_OVOL2 skin_parameters[29]
#define CMYC_pmax_OVOL2 skin_parameters[30]
#define CMYC_c_OVOL2 skin_parameters[31]
#define CMYC_h_OVOL2 skin_parameters[32]

// linear interactions
#define OVOL1_decay skin_parameters[33]
#define OVOL2_decay skin_parameters[34]
#define CMYC_decay skin_parameters[35]

// species index
#define NOTCH_species 0
#define DELTA_species 1
#define BOUND_species 2
#define BMA_species 3
#define OVOL1_species 4
#define OVOL2_species 5
#define CMYC_species 6

// boundary conditions
//#define PERIODIC_BOUNDARY 0
#define PERIODIC_BOUNDARY 1
//#define HARD_Z 0
#define HARD_Z 1

// For 128 and 250 cells
#define BOUNDARY_X 5.9
#define BOUNDARY_Y 5.9
// For 500 cells
//#define BOUNDARY_X 9.9
//#define BOUNDARY_Y 9.9
// For 1000 cells
//#define BOUNDARY_X 14.9
//#define BOUNDARY_Y 14.9
// For 5000 cells
//#define BOUNDARY_X 29.9
//#define BOUNDARY_Y 29.9

#define BOUNDARY_Z 10.0


//
////////////////////////////////////////////////////////////////////////////////////
//


#define dist(X1, Y1, Z1, X2, Y2, Z2) sqrt( ((X1) - (X2))*((X1) - (X2)) + ((Y1) - (Y2))*((Y1) - (Y2)) + ((Z1) - (Z2))*((Z1) - (Z2)) )

#define MORSE(r, U, UE, V, VE) (U)*exp(-(r)/(UE)) - (V)*exp(-(r)/(VE))

// only repulsive force, zero out attractive force
__device__ float P_MORSE(float r, float U, float UE, float V, float VE)
{
  float m = (U)*exp(-(r)/(UE)) - (V)*exp(-(r)/(VE));
  if (m < 0) return 0.0;
  else return m;
}

//
// Sector neighbor function:
// A given sector has 26 neighboring sectors in 3D space.
// The X dimension is considered left/right.
// The Y dimension is considered up/down.
// The Z dimension is considered front/back.
// We number them sequentially, so we can easily iterate through them
// First 2 are neighbors in 1D
#define S_X_TOT 3
// First 8 are neighbors in 2D
#define S_XY_TOT 9
// All 26 are neighbors in 3D
#define S_XYZ_TOT 27
//
#define S_NONE -1
#define S_CURRENT 0
#define S_START 1
#define S_LEFT 1
#define S_RIGHT 2
#define S_UP 3
#define S_DOWN 4
#define S_LEFT_UP 5
#define S_RIGHT_UP 6
#define S_LEFT_DOWN 7
#define S_RIGHT_DOWN 8
#define S_FRONT 9
#define S_BACK 10
#define S_LEFT_FRONT 11
#define S_RIGHT_FRONT 12
#define S_LEFT_BACK 13
#define S_RIGHT_BACK 14
#define S_UP_FRONT 15
#define S_DOWN_FRONT 16
#define S_UP_BACK 17
#define S_DOWN_BACK 18
#define S_LEFT_UP_FRONT 19
#define S_RIGHT_UP_FRONT 20
#define S_LEFT_DOWN_FRONT 21
#define S_RIGHT_DOWN_FRONT 22
#define S_LEFT_UP_BACK 23
#define S_RIGHT_UP_BACK 24
#define S_LEFT_DOWN_BACK 25
#define S_RIGHT_DOWN_BACK 26

// translate coordinates to specified mirror
__device__ void mirror_translate(float *X, float *Y, float *Z, int direction)
{
  switch (direction) {
  case S_LEFT:
    *X = *X - BOUNDARY_X; break;
  case S_RIGHT:
    *X = *X + BOUNDARY_X; break;
  case S_UP:
    *Y = *Y + BOUNDARY_Y; break;
  case S_DOWN:
    *Y = *Y - BOUNDARY_Y; break;
  case S_LEFT_UP:
    *X = *X - BOUNDARY_X; *Y = *Y + BOUNDARY_Y; break;
  case S_RIGHT_UP:
    *X = *X + BOUNDARY_X; *Y = *Y + BOUNDARY_Y; break;
  case S_LEFT_DOWN:
    *X = *X - BOUNDARY_X; *Y = *Y - BOUNDARY_Y; break;
  case S_RIGHT_DOWN:
    *X = *X + BOUNDARY_X; *Y = *Y - BOUNDARY_Y; break;
  case S_FRONT:
    *Z = *Z - BOUNDARY_Z; break;
  case S_BACK:
    *Z = *Z + BOUNDARY_Z; break;
  case S_LEFT_FRONT:
    *X = *X - BOUNDARY_X; *Z = *Z - BOUNDARY_Z; break;
  case S_RIGHT_FRONT:
    *X = *X + BOUNDARY_X; *Z = *Z - BOUNDARY_Z; break;
  case S_LEFT_BACK:
    *X = *X - BOUNDARY_X; *Z = *Z + BOUNDARY_Z; break;
  case S_RIGHT_BACK:
    *X = *X + BOUNDARY_X; *Z = *Z + BOUNDARY_Z; break;
  case S_UP_FRONT:
    *Y = *Y + BOUNDARY_Y; *Z = *Z - BOUNDARY_Z; break;
  case S_DOWN_FRONT:
    *Y = *Y - BOUNDARY_Y; *Z = *Z - BOUNDARY_Z; break;
  case S_UP_BACK:
    *Y = *Y + BOUNDARY_Y; *Z = *Z + BOUNDARY_Z; break;
  case S_DOWN_BACK:
    *Y = *Y - BOUNDARY_Y; *Z = *Z + BOUNDARY_Z; break;
  case S_LEFT_UP_FRONT:
    *X = *X - BOUNDARY_X; *Y = *Y + BOUNDARY_Y; *Z = *Z - BOUNDARY_Z; break;
  case S_RIGHT_UP_FRONT:
    *X = *X + BOUNDARY_X; *Y = *Y + BOUNDARY_Y; *Z = *Z - BOUNDARY_Z; break;
  case S_LEFT_DOWN_FRONT:
    *X = *X - BOUNDARY_X; *Y = *Y - BOUNDARY_Y; *Z = *Z - BOUNDARY_Z; break;
  case S_RIGHT_DOWN_FRONT:
    *X = *X + BOUNDARY_X; *Y = *Y - BOUNDARY_Y; *Z = *Z - BOUNDARY_Z; break;
  case S_LEFT_UP_BACK:
    *X = *X - BOUNDARY_X; *Y = *Y + BOUNDARY_Y; *Z = *Z + BOUNDARY_Z; break;
  case S_RIGHT_UP_BACK:
    *X = *X + BOUNDARY_X; *Y = *Y + BOUNDARY_Y; *Z = *Z + BOUNDARY_Z; break;
  case S_LEFT_DOWN_BACK:
    *X = *X - BOUNDARY_X; *Y = *Y - BOUNDARY_Y; *Z = *Z + BOUNDARY_Z; break;
  case S_RIGHT_DOWN_BACK:
    *X = *X + BOUNDARY_X; *Y = *Y - BOUNDARY_Y; *Z = *Z + BOUNDARY_Z; break;
  }
}

// intracellular mirror calculations for periodic boundary conditions
__device__ void
intracellular_mirror(int elemNum, int cellNum, int pitch, int k, float *intraX, float *intraY, float *intraZ, float *X, float *Y, float *Z)
{
  float oX, oY, r, V;

  // left mirror
  oX = X[k*pitch+cellNum] - BOUNDARY_X;
  r = dist(X[elemNum*pitch+cellNum], Y[elemNum*pitch+cellNum], Z[elemNum*pitch+cellNum], oX, Y[k*pitch+cellNum], Z[k*pitch+cellNum]);
  V = MORSE(r, INTRA_U0, INTRA_ETA0, INTRA_U1, INTRA_ETA1);
  *intraX += V * (X[elemNum*pitch+cellNum] - oX);
  *intraY += V * (Y[elemNum*pitch+cellNum] - Y[k*pitch+cellNum]);
  *intraZ += V * (Z[elemNum*pitch+cellNum] - Z[k*pitch+cellNum]);

  // right mirror
  oX = X[k*pitch+cellNum] + BOUNDARY_X;
  r = dist(X[elemNum*pitch+cellNum], Y[elemNum*pitch+cellNum], Z[elemNum*pitch+cellNum], oX, Y[k*pitch+cellNum], Z[k*pitch+cellNum]);
  V = MORSE(r, INTRA_U0, INTRA_ETA0, INTRA_U1, INTRA_ETA1);
  *intraX += V * (X[elemNum*pitch+cellNum] - oX);
  *intraY += V * (Y[elemNum*pitch+cellNum] - Y[k*pitch+cellNum]);
  *intraZ += V * (Z[elemNum*pitch+cellNum] - Z[k*pitch+cellNum]);

  // down mirror
  oY = Y[k*pitch+cellNum] - BOUNDARY_Y;
  r = dist(X[elemNum*pitch+cellNum], Y[elemNum*pitch+cellNum], Z[elemNum*pitch+cellNum], X[k*pitch+cellNum], oY, Z[k*pitch+cellNum]);
  V = MORSE(r, INTRA_U0, INTRA_ETA0, INTRA_U1, INTRA_ETA1);
  *intraX += V * (X[elemNum*pitch+cellNum] - X[k*pitch+cellNum]);
  *intraY += V * (Y[elemNum*pitch+cellNum] - oY);
  *intraZ += V * (Z[elemNum*pitch+cellNum] - Z[k*pitch+cellNum]);

  // up mirror
  oY = Y[k*pitch+cellNum] + BOUNDARY_Y;
  r = dist(X[elemNum*pitch+cellNum], Y[elemNum*pitch+cellNum], Z[elemNum*pitch+cellNum], X[k*pitch+cellNum], oY, Z[k*pitch+cellNum]);
  V = MORSE(r, INTRA_U0, INTRA_ETA0, INTRA_U1, INTRA_ETA1);
  *intraX += V * (X[elemNum*pitch+cellNum] - X[k*pitch+cellNum]);
  *intraY += V * (Y[elemNum*pitch+cellNum] - oY);
  *intraZ += V * (Z[elemNum*pitch+cellNum] - Z[k*pitch+cellNum]);

  // up-left mirror
  oX = X[k*pitch+cellNum] - BOUNDARY_X;
  oY = Y[k*pitch+cellNum] + BOUNDARY_Y;
  r = dist(X[elemNum*pitch+cellNum], Y[elemNum*pitch+cellNum], Z[elemNum*pitch+cellNum], oX, oY, Z[k*pitch+cellNum]);
  V = MORSE(r, INTRA_U0, INTRA_ETA0, INTRA_U1, INTRA_ETA1);
  *intraX += V * (X[elemNum*pitch+cellNum] - oX);
  *intraY += V * (Y[elemNum*pitch+cellNum] - oY);
  *intraZ += V * (Z[elemNum*pitch+cellNum] - Z[k*pitch+cellNum]);

  // up-right mirror
  oX = X[k*pitch+cellNum] + BOUNDARY_X;
  oY = Y[k*pitch+cellNum] + BOUNDARY_Y;
  r = dist(X[elemNum*pitch+cellNum], Y[elemNum*pitch+cellNum], Z[elemNum*pitch+cellNum], oX, oY, Z[k*pitch+cellNum]);
  V = MORSE(r, INTRA_U0, INTRA_ETA0, INTRA_U1, INTRA_ETA1);
  *intraX += V * (X[elemNum*pitch+cellNum] - oX);
  *intraY += V * (Y[elemNum*pitch+cellNum] - oY);
  *intraZ += V * (Z[elemNum*pitch+cellNum] - Z[k*pitch+cellNum]);

  // down-left mirror
  oX = X[k*pitch+cellNum] - BOUNDARY_X;
  oY = Y[k*pitch+cellNum] - BOUNDARY_Y;
  r = dist(X[elemNum*pitch+cellNum], Y[elemNum*pitch+cellNum], Z[elemNum*pitch+cellNum], oX, oY, Z[k*pitch+cellNum]);
  V = MORSE(r, INTRA_U0, INTRA_ETA0, INTRA_U1, INTRA_ETA1);
  *intraX += V * (X[elemNum*pitch+cellNum] - oX);
  *intraY += V * (Y[elemNum*pitch+cellNum] - oY);
  *intraZ += V * (Z[elemNum*pitch+cellNum] - Z[k*pitch+cellNum]);

  // down-right mirror
  oX = X[k*pitch+cellNum] + BOUNDARY_X;
  oY = Y[k*pitch+cellNum] - BOUNDARY_Y;
  r = dist(X[elemNum*pitch+cellNum], Y[elemNum*pitch+cellNum], Z[elemNum*pitch+cellNum], oX, oY, Z[k*pitch+cellNum]);
  V = MORSE(r, INTRA_U0, INTRA_ETA0, INTRA_U1, INTRA_ETA1);
  *intraX += V * (X[elemNum*pitch+cellNum] - oX);
  *intraY += V * (Y[elemNum*pitch+cellNum] - oY);
  *intraZ += V * (Z[elemNum*pitch+cellNum] - Z[k*pitch+cellNum]);
}

// intercellular mirror calculations for periodic boundary conditions
__device__ void
intercellular_mirror(int elemNum, int cellNum, int pitch, int k, int j, float *interX, float *interY, float *interZ, float *X, float *Y, float *Z)
{
  float oX, oY, r, V;

  // left mirror
  oX = X[k*pitch+j] - BOUNDARY_X;
  r = dist(X[elemNum*pitch+cellNum], Y[elemNum*pitch+cellNum], Z[elemNum*pitch+cellNum], oX, Y[k*pitch+j], Z[k*pitch+j]);
  if (r <= INTER_DIST) {
    V = P_MORSE(r, INTER_U0, INTER_ETA0, INTER_U1, INTER_ETA1);
    *interX += V * (X[elemNum*pitch+cellNum] - oX);
    *interY += V * (Y[elemNum*pitch+cellNum] - Y[k*pitch+j]);
    *interZ += V * (Z[elemNum*pitch+cellNum] - Z[k*pitch+j]);
  }

  // right mirror
  oX = X[k*pitch+j] + BOUNDARY_X;
  r = dist(X[elemNum*pitch+cellNum], Y[elemNum*pitch+cellNum], Z[elemNum*pitch+cellNum], oX, Y[k*pitch+j], Z[k*pitch+j]);
  if (r <= INTER_DIST) {
    V = P_MORSE(r, INTER_U0, INTER_ETA0, INTER_U1, INTER_ETA1);
    *interX += V * (X[elemNum*pitch+cellNum] - oX);
    *interY += V * (Y[elemNum*pitch+cellNum] - Y[k*pitch+j]);
    *interZ += V * (Z[elemNum*pitch+cellNum] - Z[k*pitch+j]);
  }

  // down mirror
  oY = Y[k*pitch+j] - BOUNDARY_Y;
  r = dist(X[elemNum*pitch+cellNum], Y[elemNum*pitch+cellNum], Z[elemNum*pitch+cellNum], X[k*pitch+j], oY, Z[k*pitch+j]);
  if (r <= INTER_DIST) {
    V = P_MORSE(r, INTER_U0, INTER_ETA0, INTER_U1, INTER_ETA1);
    *interX += V * (X[elemNum*pitch+cellNum] - X[k*pitch+j]);
    *interY += V * (Y[elemNum*pitch+cellNum] - oY);
    *interZ += V * (Z[elemNum*pitch+cellNum] - Z[k*pitch+j]);
  }

  // up mirror
  oY = Y[k*pitch+j] + BOUNDARY_Y;
  r = dist(X[elemNum*pitch+cellNum], Y[elemNum*pitch+cellNum], Z[elemNum*pitch+cellNum], X[k*pitch+j], oY, Z[k*pitch+j]);
  if (r <= INTER_DIST) {
    V = P_MORSE(r, INTER_U0, INTER_ETA0, INTER_U1, INTER_ETA1);
    *interX += V * (X[elemNum*pitch+cellNum] - X[k*pitch+j]);
    *interY += V * (Y[elemNum*pitch+cellNum] - oY);
    *interZ += V * (Z[elemNum*pitch+cellNum] - Z[k*pitch+j]);
  }

  // up-left mirror
  oX = X[k*pitch+j] - BOUNDARY_X;
  oY = Y[k*pitch+j] + BOUNDARY_Y;
  r = dist(X[elemNum*pitch+cellNum], Y[elemNum*pitch+cellNum], Z[elemNum*pitch+cellNum], oX, oY, Z[k*pitch+j]);
  if (r <= INTER_DIST) {
    V = P_MORSE(r, INTER_U0, INTER_ETA0, INTER_U1, INTER_ETA1);
    *interX += V * (X[elemNum*pitch+cellNum] - oX);
    *interY += V * (Y[elemNum*pitch+cellNum] - oY);
    *interZ += V * (Z[elemNum*pitch+cellNum] - Z[k*pitch+j]);
  }

  // up-right mirror
  oX = X[k*pitch+j] + BOUNDARY_X;
  oY = Y[k*pitch+j] + BOUNDARY_Y;
  r = dist(X[elemNum*pitch+cellNum], Y[elemNum*pitch+cellNum], Z[elemNum*pitch+cellNum], oX, oY, Z[k*pitch+j]);
  if (r <= INTER_DIST) {
    V = P_MORSE(r, INTER_U0, INTER_ETA0, INTER_U1, INTER_ETA1);
    *interX += V * (X[elemNum*pitch+cellNum] - oX);
    *interY += V * (Y[elemNum*pitch+cellNum] - oY);
    *interZ += V * (Z[elemNum*pitch+cellNum] - Z[k*pitch+j]);
  }

  // down-left mirror
  oX = X[k*pitch+j] - BOUNDARY_X;
  oY = Y[k*pitch+j] - BOUNDARY_Y;
  r = dist(X[elemNum*pitch+cellNum], Y[elemNum*pitch+cellNum], Z[elemNum*pitch+cellNum], oX, oY, Z[k*pitch+j]);
  if (r <= INTER_DIST) {
    V = P_MORSE(r, INTER_U0, INTER_ETA0, INTER_U1, INTER_ETA1);
    *interX += V * (X[elemNum*pitch+cellNum] - oX);
    *interY += V * (Y[elemNum*pitch+cellNum] - oY);
    *interZ += V * (Z[elemNum*pitch+cellNum] - Z[k*pitch+j]);
  }

  // down-right mirror
  oX = X[k*pitch+j] + BOUNDARY_X;
  oY = Y[k*pitch+j] - BOUNDARY_Y;
  r = dist(X[elemNum*pitch+cellNum], Y[elemNum*pitch+cellNum], Z[elemNum*pitch+cellNum], oX, oY, Z[k*pitch+j]);
  if (r <= INTER_DIST) {
    V = P_MORSE(r, INTER_U0, INTER_ETA0, INTER_U1, INTER_ETA1);
    *interX += V * (X[elemNum*pitch+cellNum] - oX);
    *interY += V * (Y[elemNum*pitch+cellNum] - oY);
    *interZ += V * (Z[elemNum*pitch+cellNum] - Z[k*pitch+j]);
  }
}

__device__ int
distance_check(float X, float Y, float Z, float oX, float oY, float oZ, int maxDirection, float distance)
{
  int direction;
  float nX, nY, nZ;

  for (direction = S_CURRENT; direction < maxDirection; ++direction) {
    nX = oX;
    nY = oY;
    nZ = oZ;
    mirror_translate(&nX, &nY, &nZ, direction);
    if (dist(X, Y, Z, nX, nY, nZ) <= distance)
      return direction;
  }

  return S_NONE;
}

//
//
//
////////////////////////////////////////////////////////////////////////////////////


//
// Cell movement, naive implementation
//

// F1 kernel for 2nd order Runge-Kutta
__global__ void
sem_kernel_F1(int *numOfElements, float *X, float *X_F1, float *X_F2, float *Y, float *Y_F1, float *Y_F2, float *Z, float *Z_F1, float *Z_F2,
	      int *elementType, size_t pitch, int numOfCells, int maxCells, int maxElements, float dt)
{
   int cellNum = blockIdx.x * blockDim.x + threadIdx.x;
   int elemNum = blockIdx.y * blockDim.y + threadIdx.y;

   if (cellNum >= numOfCells) return;
   if (elemNum >= numOfElements[cellNum]) return;

   // intracellular
   float intraX = 0.0;
   float intraY = 0.0;
   float intraZ = 0.0;

   int j, k;
   float r, V;
   for (k = 0; k < numOfElements[cellNum]; ++k) {
     if (k == elemNum) continue;

     r = dist(X[elemNum*pitch+cellNum], Y[elemNum*pitch+cellNum], Z[elemNum*pitch+cellNum],
	      X[k*pitch+cellNum], Y[k*pitch+cellNum], Z[k*pitch+cellNum]);
     V = MORSE(r, INTRA_U0, INTRA_ETA0, INTRA_U1, INTRA_ETA1);
     intraX += V * (X[elemNum*pitch+cellNum] - X[k*pitch+cellNum]);
     intraY += V * (Y[elemNum*pitch+cellNum] - Y[k*pitch+cellNum]);
     intraZ += V * (Z[elemNum*pitch+cellNum] - Z[k*pitch+cellNum]);

#if PERIODIC_BOUNDARY
     intracellular_mirror(elemNum, cellNum, pitch, k, &intraX, &intraY, &intraZ, X, Y, Z);
#endif
   }

   // intercellular
   float interX = 0.0;
   float interY = 0.0;
   float interZ = 0.0;

   for (j = 0; j < numOfCells; ++j) {
     if (j == cellNum) continue;

     for (k = 0; k < numOfElements[j]; ++k) {

       r = dist(X[elemNum*pitch+cellNum], Y[elemNum*pitch+cellNum], Z[elemNum*pitch+cellNum], X[k*pitch+j], Y[k*pitch+j], Z[k*pitch+j]);
       if (r <= INTER_DIST) {
	 V = P_MORSE(r, INTER_U0, INTER_ETA0, INTER_U1, INTER_ETA1);
	 interX += V * (X[elemNum*pitch+cellNum] - X[k*pitch+j]);
	 interY += V * (Y[elemNum*pitch+cellNum] - Y[k*pitch+j]);
	 interZ += V * (Z[elemNum*pitch+cellNum] - Z[k*pitch+j]);
       }

#if PERIODIC_BOUNDARY
       intercellular_mirror(elemNum, cellNum, pitch, k, j, &interX, &interY, &interZ, X, Y, Z);
#endif
     }
   }

   // basement membrane
   if (elementType[elemNum*pitch+cellNum] == 1) {
     r = dist(X[elemNum*pitch+cellNum], Y[elemNum*pitch+cellNum], Z[elemNum*pitch+cellNum],
	      X[elemNum*pitch+cellNum], Y[elemNum*pitch+cellNum], 0);
     V = MORSE(r, INTRA_U0, INTRA_ETA0, INTRA_U1, INTRA_ETA1);
     interZ += V * (Z[elemNum*pitch+cellNum] - 0);
   }

   // update
   X_F1[elemNum*pitch+cellNum] = X[elemNum*pitch+cellNum] + 0.5 * dt * (intraX + interX);
   Y_F1[elemNum*pitch+cellNum] = Y[elemNum*pitch+cellNum] + 0.5 * dt * (intraY + interY);
   Z_F1[elemNum*pitch+cellNum] = Z[elemNum*pitch+cellNum] + 0.5 * dt * (intraZ + interZ);

#if PERIODIC_BOUNDARY
   if (X_F1[elemNum*pitch+cellNum] < 0.0) X_F1[elemNum*pitch+cellNum] = X_F1[elemNum*pitch+cellNum] + BOUNDARY_X;
   if (X_F1[elemNum*pitch+cellNum] > BOUNDARY_X) X_F1[elemNum*pitch+cellNum] = X_F1[elemNum*pitch+cellNum] - BOUNDARY_X;
   if (Y_F1[elemNum*pitch+cellNum] < 0.0) Y_F1[elemNum*pitch+cellNum] = Y_F1[elemNum*pitch+cellNum] + BOUNDARY_Y;
   if (Y_F1[elemNum*pitch+cellNum] > BOUNDARY_Y) Y_F1[elemNum*pitch+cellNum] = Y_F1[elemNum*pitch+cellNum] - BOUNDARY_Y;
   //if (Z_F1[elemNum*pitch+cellNum] < 0.0) Z_F1[elemNum*pitch+cellNum] = Z_F1[elemNum*pitch+cellNum] + BOUNDARY_Z;
   //if (Z_F1[elemNum*pitch+cellNum] > BOUNDARY_Z) Z_F1[elemNum*pitch+cellNum] = Z_F1[elemNum*pitch+cellNum] - BOUNDARY_Z;
#endif

#if HARD_Z
   if (Z_F1[elemNum*pitch+cellNum] < 0.0) Z_F1[elemNum*pitch+cellNum] = 0.0;
   if (Z_F1[elemNum*pitch+cellNum] > BOUNDARY_Z) Z_F1[elemNum*pitch+cellNum] = BOUNDARY_Z;
#endif
}

// F2 kernel for 2nd order Runge-Kutta
__global__ void
sem_kernel_F2(int *numOfElements, float *X, float *X_F1, float *X_F2, float *Y, float *Y_F1, float *Y_F2, float *Z, float *Z_F1, float *Z_F2,
	      int *elementType, size_t pitch, int numOfCells, int maxCells, int maxElements, float dt)
{
   int cellNum = blockIdx.x * blockDim.x + threadIdx.x;
   int elemNum = blockIdx.y * blockDim.y + threadIdx.y;

   if (cellNum >= numOfCells) return;
   if (elemNum >= numOfElements[cellNum]) return;

   // intracellular
   float intraX = 0.0;
   float intraY = 0.0;
   float intraZ = 0.0;

   int j, k;
   float r, V;
   for (k = 0; k < numOfElements[cellNum]; ++k) {
     if (k == elemNum) continue;

     r = dist(X_F1[elemNum*pitch+cellNum], Y_F1[elemNum*pitch+cellNum], Z_F1[elemNum*pitch+cellNum],
	      X_F1[k*pitch+cellNum], Y_F1[k*pitch+cellNum], Z_F1[k*pitch+cellNum]);
     V = MORSE(r, INTRA_U0, INTRA_ETA0, INTRA_U1, INTRA_ETA1);
     intraX += V * (X_F1[elemNum*pitch+cellNum] - X_F1[k*pitch+cellNum]);
     intraY += V * (Y_F1[elemNum*pitch+cellNum] - Y_F1[k*pitch+cellNum]);
     intraZ += V * (Z_F1[elemNum*pitch+cellNum] - Z_F1[k*pitch+cellNum]);

#if PERIODIC_BOUNDARY
     intracellular_mirror(elemNum, cellNum, pitch, k, &intraX, &intraY, &intraZ, X_F1, Y_F1, Z_F1);
#endif
   }

   // intercellular
   float interX = 0.0;
   float interY = 0.0;
   float interZ = 0.0;

   for (j = 0; j < numOfCells; ++j) {
     if (j == cellNum) continue;

     for (k = 0; k < numOfElements[j]; ++k) {

       r = dist(X_F1[elemNum*pitch+cellNum], Y_F1[elemNum*pitch+cellNum], Z_F1[elemNum*pitch+cellNum], X_F1[k*pitch+j], Y_F1[k*pitch+j], Z_F1[k*pitch+j]);
       if (r <= INTER_DIST) {
	 V = P_MORSE(r, INTER_U0, INTER_ETA0, INTER_U1, INTER_ETA1);
	 interX += V * (X_F1[elemNum*pitch+cellNum] - X_F1[k*pitch+j]);
	 interY += V * (Y_F1[elemNum*pitch+cellNum] - Y_F1[k*pitch+j]);
	 interZ += V * (Z_F1[elemNum*pitch+cellNum] - Z_F1[k*pitch+j]);
       }

#if PERIODIC_BOUNDARY
       intercellular_mirror(elemNum, cellNum, pitch, k, j, &interX, &interY, &interZ, X_F1, Y_F1, Z_F1);
#endif
     }
   }

   // basement membrane
   if (elementType[elemNum*pitch+cellNum] == 1) {
     r = dist(X[elemNum*pitch+cellNum], Y[elemNum*pitch+cellNum], Z[elemNum*pitch+cellNum],
	      X[elemNum*pitch+cellNum], Y[elemNum*pitch+cellNum], 0);
     V = MORSE(r, INTRA_U0, INTRA_ETA0, INTRA_U1, INTRA_ETA1);
     interZ += V * (Z[elemNum*pitch+cellNum] - 0);
   }

   // update
   X_F2[elemNum*pitch+cellNum] = X[elemNum*pitch+cellNum] + dt * (intraX + interX);
   Y_F2[elemNum*pitch+cellNum] = Y[elemNum*pitch+cellNum] + dt * (intraY + interY);
   Z_F2[elemNum*pitch+cellNum] = Z[elemNum*pitch+cellNum] + dt * (intraZ + interZ);

#if PERIODIC_BOUNDARY
   if (X_F2[elemNum*pitch+cellNum] < 0.0) X_F2[elemNum*pitch+cellNum] = X_F2[elemNum*pitch+cellNum] + BOUNDARY_X;
   if (X_F2[elemNum*pitch+cellNum] > BOUNDARY_X) X_F2[elemNum*pitch+cellNum] = X_F2[elemNum*pitch+cellNum] - BOUNDARY_X;
   if (Y_F2[elemNum*pitch+cellNum] < 0.0) Y_F2[elemNum*pitch+cellNum] = Y_F2[elemNum*pitch+cellNum] + BOUNDARY_Y;
   if (Y_F2[elemNum*pitch+cellNum] > BOUNDARY_Y) Y_F2[elemNum*pitch+cellNum] = Y_F2[elemNum*pitch+cellNum] - BOUNDARY_Y;
   //if (Z_F2[elemNum*pitch+cellNum] < 0.0) Z_F2[elemNum*pitch+cellNum] = Z_F2[elemNum*pitch+cellNum] + BOUNDARY_Z;
   //if (Z_F2[elemNum*pitch+cellNum] > BOUNDARY_Z) Z_F2[elemNum*pitch+cellNum] = Z_F2[elemNum*pitch+cellNum] - BOUNDARY_Z;
#endif

#if HARD_Z
   if (Z_F2[elemNum*pitch+cellNum] < 0.0) Z_F2[elemNum*pitch+cellNum] = 0.0;
   if (Z_F2[elemNum*pitch+cellNum] > BOUNDARY_Z) Z_F2[elemNum*pitch+cellNum] = BOUNDARY_Z;
#endif
}


//
////////////////////////////////////////////////////////////////////////////////////
//

// Calculate cell center
__global__ void
skin_center_kernel(int *numOfElements, float *X, float *Y, float *Z, int *elementType,
		   float *cellCenterX, float *cellCenterY, float *cellCenterZ, size_t pitch, float *speciesData,
		   int numOfCells, int maxCells, int maxElements, float dt)
{
  int cellNum = blockIdx.x * blockDim.x + threadIdx.x;
  int elemNum;
  
  if (cellNum >= numOfCells) return;
  
  float cX = 0.0;
  float cY = 0.0;
  float cZ = 0.0;
  float minX, maxX;
  float minY, maxY;
  
  minX = X[cellNum];
  maxX = X[cellNum];
  minY = Y[cellNum];
  maxY = Y[cellNum];
  speciesData[BMA_species*pitch+cellNum] = 0.0;
  for (elemNum = 0; elemNum < numOfElements[cellNum]; ++elemNum) {
    cX += X[elemNum*pitch+cellNum];
    cY += Y[elemNum*pitch+cellNum];
    cZ += Z[elemNum*pitch+cellNum];
    if (X[elemNum*pitch+cellNum] < minX) minX = X[elemNum*pitch+cellNum];
    if (X[elemNum*pitch+cellNum] > maxX) maxX = X[elemNum*pitch+cellNum];
    if (Y[elemNum*pitch+cellNum] < minY) minY = Y[elemNum*pitch+cellNum];
    if (Y[elemNum*pitch+cellNum] > maxY) maxY = Y[elemNum*pitch+cellNum];

    // calculate potential with basement membrane
    if (elementType[elemNum*pitch+cellNum] == 1) {
      float r = dist(X[elemNum*pitch+cellNum], Y[elemNum*pitch+cellNum], Z[elemNum*pitch+cellNum],
		     X[elemNum*pitch+cellNum], Y[elemNum*pitch+cellNum], 0);
      speciesData[BMA_species*pitch+cellNum] += r;
    }
  }
  
  cX = cX / (float)numOfElements[cellNum];
  cY = cY / (float)numOfElements[cellNum];
  cZ = cZ / (float)numOfElements[cellNum];
  
  // handle special case when cell is split across periodic boundary
  if ((maxX - minX) > (BOUNDARY_X / 2)) {
    cX = 0;
    for (elemNum = 0; elemNum < numOfElements[cellNum]; ++elemNum) {
      if (X[elemNum*pitch+cellNum] > (BOUNDARY_X / 2))
        cX += X[elemNum*pitch+cellNum] - BOUNDARY_X;
      else
        cX += X[elemNum*pitch+cellNum];
    }
    cX = cX / (float)numOfElements[cellNum];
    if (cX < 0) cX += BOUNDARY_X;
  }
  if ((maxY - minY) > (BOUNDARY_Y / 2)) {
    cY = 0;
    for (elemNum = 0; elemNum < numOfElements[cellNum]; ++elemNum) {
      if (Y[elemNum*pitch+cellNum] > (BOUNDARY_Y / 2))
        cY += Y[elemNum*pitch+cellNum] - BOUNDARY_Y;
      else
        cY += Y[elemNum*pitch+cellNum];
    }
    cY = cY / (float)numOfElements[cellNum];
    if (cY < 0) cY += BOUNDARY_Y;
  }
  
  cellCenterX[cellNum] = cX;
  cellCenterY[cellNum] = cY;
  cellCenterZ[cellNum] = cZ;  

  if (speciesData[BMA_species*pitch+cellNum] != 0.0) speciesData[BMA_species*pitch+cellNum] = 1.0 / speciesData[BMA_species*pitch+cellNum];
}


//
// Cell movement, cell center implementation
//

// F1 kernel for 2nd order Runge-Kutta
__global__ void
skin_moveKernel_F1(int *numOfElements, float *X, float *X_F1, float *X_F2, float *Y, float *Y_F1, float *Y_F2, float *Z, float *Z_F1, float *Z_F2,
		   float *cX, float *cY, float *cZ,
		   int *elementType, size_t pitch, int numOfCells, int maxCells, int maxElements, float dt)
{
   int cellNum = blockIdx.x * blockDim.x + threadIdx.x;
   int elemNum = blockIdx.y * blockDim.y + threadIdx.y;

   if (cellNum >= numOfCells) return;
   if (elemNum >= numOfElements[cellNum]) return;

   // intracellular
   float intraX = 0.0;
   float intraY = 0.0;
   float intraZ = 0.0;

   int j, k;
   float r, V;
   for (k = 0; k < numOfElements[cellNum]; ++k) {
     if (k == elemNum) continue;

     r = dist(X[elemNum*pitch+cellNum], Y[elemNum*pitch+cellNum], Z[elemNum*pitch+cellNum],
	      X[k*pitch+cellNum], Y[k*pitch+cellNum], Z[k*pitch+cellNum]);
     V = MORSE(r, INTRA_U0, INTRA_ETA0, INTRA_U1, INTRA_ETA1);
     intraX += V * (X[elemNum*pitch+cellNum] - X[k*pitch+cellNum]);
     intraY += V * (Y[elemNum*pitch+cellNum] - Y[k*pitch+cellNum]);
     intraZ += V * (Z[elemNum*pitch+cellNum] - Z[k*pitch+cellNum]);

#if PERIODIC_BOUNDARY
     intracellular_mirror(elemNum, cellNum, pitch, k, &intraX, &intraY, &intraZ, X, Y, Z);
#endif
   }

   // intercellular
   float interX = 0.0;
   float interY = 0.0;
   float interZ = 0.0;

   for (j = 0; j < numOfCells; ++j) {
     if (j == cellNum) continue;

     r = dist(cX[cellNum], cY[cellNum], cZ[cellNum], cX[j], cY[j], cZ[j]);
     if (r > INTER_DIST) continue;

     for (k = 0; k < numOfElements[j]; ++k) {

       r = dist(X[elemNum*pitch+cellNum], Y[elemNum*pitch+cellNum], Z[elemNum*pitch+cellNum], X[k*pitch+j], Y[k*pitch+j], Z[k*pitch+j]);
       if (r <= INTER_DIST) {
	 V = P_MORSE(r, INTER_U0, INTER_ETA0, INTER_U1, INTER_ETA1);
	 interX += V * (X[elemNum*pitch+cellNum] - X[k*pitch+j]);
	 interY += V * (Y[elemNum*pitch+cellNum] - Y[k*pitch+j]);
	 interZ += V * (Z[elemNum*pitch+cellNum] - Z[k*pitch+j]);
       }

#if PERIODIC_BOUNDARY
       intercellular_mirror(elemNum, cellNum, pitch, k, j, &interX, &interY, &interZ, X, Y, Z);
#endif
     }
   }

   // basement membrane
   if (elementType[elemNum*pitch+cellNum] == 1) {
     r = dist(X[elemNum*pitch+cellNum], Y[elemNum*pitch+cellNum], Z[elemNum*pitch+cellNum],
	      X[elemNum*pitch+cellNum], Y[elemNum*pitch+cellNum], 0);
     V = MORSE(r, INTRA_U0, INTRA_ETA0, INTRA_U1, INTRA_ETA1);
     interZ += V * (Z[elemNum*pitch+cellNum] - 0);
   }

   // update
   X_F1[elemNum*pitch+cellNum] = X[elemNum*pitch+cellNum] + 0.5 * dt * (intraX + interX);
   Y_F1[elemNum*pitch+cellNum] = Y[elemNum*pitch+cellNum] + 0.5 * dt * (intraY + interY);
   Z_F1[elemNum*pitch+cellNum] = Z[elemNum*pitch+cellNum] + 0.5 * dt * (intraZ + interZ);

#if PERIODIC_BOUNDARY
   if (X_F1[elemNum*pitch+cellNum] < 0.0) X_F1[elemNum*pitch+cellNum] = X_F1[elemNum*pitch+cellNum] + BOUNDARY_X;
   if (X_F1[elemNum*pitch+cellNum] > BOUNDARY_X) X_F1[elemNum*pitch+cellNum] = X_F1[elemNum*pitch+cellNum] - BOUNDARY_X;
   if (Y_F1[elemNum*pitch+cellNum] < 0.0) Y_F1[elemNum*pitch+cellNum] = Y_F1[elemNum*pitch+cellNum] + BOUNDARY_Y;
   if (Y_F1[elemNum*pitch+cellNum] > BOUNDARY_Y) Y_F1[elemNum*pitch+cellNum] = Y_F1[elemNum*pitch+cellNum] - BOUNDARY_Y;
   //if (Z_F1[elemNum*pitch+cellNum] < 0.0) Z_F1[elemNum*pitch+cellNum] = Z_F1[elemNum*pitch+cellNum] + BOUNDARY_Z;
   //if (Z_F1[elemNum*pitch+cellNum] > BOUNDARY_Z) Z_F1[elemNum*pitch+cellNum] = Z_F1[elemNum*pitch+cellNum] - BOUNDARY_Z;
#endif

#if HARD_Z
   if (Z_F1[elemNum*pitch+cellNum] < 0.0) Z_F1[elemNum*pitch+cellNum] = 0.0;
   if (Z_F1[elemNum*pitch+cellNum] > BOUNDARY_Z) Z_F1[elemNum*pitch+cellNum] = BOUNDARY_Z;
#endif
}

// F2 kernel for 2nd order Runge-Kutta
__global__ void
skin_moveKernel_F2(int *numOfElements, float *X, float *X_F1, float *X_F2, float *Y, float *Y_F1, float *Y_F2, float *Z, float *Z_F1, float *Z_F2,
		   float *cX, float *cY, float *cZ,		   
		   int *elementType, size_t pitch, int numOfCells, int maxCells, int maxElements, float dt)
{
   int cellNum = blockIdx.x * blockDim.x + threadIdx.x;
   int elemNum = blockIdx.y * blockDim.y + threadIdx.y;

   if (cellNum >= numOfCells) return;
   if (elemNum >= numOfElements[cellNum]) return;

   // intracellular
   float intraX = 0.0;
   float intraY = 0.0;
   float intraZ = 0.0;

   int j, k;
   float r, V;
   for (k = 0; k < numOfElements[cellNum]; ++k) {
     if (k == elemNum) continue;

     r = dist(X_F1[elemNum*pitch+cellNum], Y_F1[elemNum*pitch+cellNum], Z_F1[elemNum*pitch+cellNum],
	      X_F1[k*pitch+cellNum], Y_F1[k*pitch+cellNum], Z_F1[k*pitch+cellNum]);
     V = MORSE(r, INTRA_U0, INTRA_ETA0, INTRA_U1, INTRA_ETA1);
     intraX += V * (X_F1[elemNum*pitch+cellNum] - X_F1[k*pitch+cellNum]);
     intraY += V * (Y_F1[elemNum*pitch+cellNum] - Y_F1[k*pitch+cellNum]);
     intraZ += V * (Z_F1[elemNum*pitch+cellNum] - Z_F1[k*pitch+cellNum]);

#if PERIODIC_BOUNDARY
     intracellular_mirror(elemNum, cellNum, pitch, k, &intraX, &intraY, &intraZ, X_F1, Y_F1, Z_F1);
#endif
   }

   // intercellular
   float interX = 0.0;
   float interY = 0.0;
   float interZ = 0.0;

   for (j = 0; j < numOfCells; ++j) {
     if (j == cellNum) continue;

     r = dist(cX[cellNum], cY[cellNum], cZ[cellNum], cX[j], cY[j], cZ[j]);
     if (r > INTER_DIST) continue;

     for (k = 0; k < numOfElements[j]; ++k) {

       r = dist(X_F1[elemNum*pitch+cellNum], Y_F1[elemNum*pitch+cellNum], Z_F1[elemNum*pitch+cellNum], X_F1[k*pitch+j], Y_F1[k*pitch+j], Z_F1[k*pitch+j]);
       if (r <= INTER_DIST) {
	 V = P_MORSE(r, INTER_U0, INTER_ETA0, INTER_U1, INTER_ETA1);
	 interX += V * (X_F1[elemNum*pitch+cellNum] - X_F1[k*pitch+j]);
	 interY += V * (Y_F1[elemNum*pitch+cellNum] - Y_F1[k*pitch+j]);
	 interZ += V * (Z_F1[elemNum*pitch+cellNum] - Z_F1[k*pitch+j]);
       }

#if PERIODIC_BOUNDARY
       intercellular_mirror(elemNum, cellNum, pitch, k, j, &interX, &interY, &interZ, X_F1, Y_F1, Z_F1);
#endif
     }
   }

   // basement membrane
   if (elementType[elemNum*pitch+cellNum] == 1) {
     r = dist(X[elemNum*pitch+cellNum], Y[elemNum*pitch+cellNum], Z[elemNum*pitch+cellNum],
	      X[elemNum*pitch+cellNum], Y[elemNum*pitch+cellNum], 0);
     V = MORSE(r, INTRA_U0, INTRA_ETA0, INTRA_U1, INTRA_ETA1);
     interZ += V * (Z[elemNum*pitch+cellNum] - 0);
   }

   // update
   X_F2[elemNum*pitch+cellNum] = X[elemNum*pitch+cellNum] + dt * (intraX + interX);
   Y_F2[elemNum*pitch+cellNum] = Y[elemNum*pitch+cellNum] + dt * (intraY + interY);
   Z_F2[elemNum*pitch+cellNum] = Z[elemNum*pitch+cellNum] + dt * (intraZ + interZ);

#if PERIODIC_BOUNDARY
   if (X_F2[elemNum*pitch+cellNum] < 0.0) X_F2[elemNum*pitch+cellNum] = X_F2[elemNum*pitch+cellNum] + BOUNDARY_X;
   if (X_F2[elemNum*pitch+cellNum] > BOUNDARY_X) X_F2[elemNum*pitch+cellNum] = X_F2[elemNum*pitch+cellNum] - BOUNDARY_X;
   if (Y_F2[elemNum*pitch+cellNum] < 0.0) Y_F2[elemNum*pitch+cellNum] = Y_F2[elemNum*pitch+cellNum] + BOUNDARY_Y;
   if (Y_F2[elemNum*pitch+cellNum] > BOUNDARY_Y) Y_F2[elemNum*pitch+cellNum] = Y_F2[elemNum*pitch+cellNum] - BOUNDARY_Y;
   //if (Z_F2[elemNum*pitch+cellNum] < 0.0) Z_F2[elemNum*pitch+cellNum] = Z_F2[elemNum*pitch+cellNum] + BOUNDARY_Z;
   //if (Z_F2[elemNum*pitch+cellNum] > BOUNDARY_Z) Z_F2[elemNum*pitch+cellNum] = Z_F2[elemNum*pitch+cellNum] - BOUNDARY_Z;
#endif

#if HARD_Z
   if (Z_F2[elemNum*pitch+cellNum] < 0.0) Z_F2[elemNum*pitch+cellNum] = 0.0;
   if (Z_F2[elemNum*pitch+cellNum] > BOUNDARY_Z) Z_F2[elemNum*pitch+cellNum] = BOUNDARY_Z;
#endif
}


//
////////////////////////////////////////////////////////////////////////////////////
//


// Average neighbor values
__global__ void
skin_neighbor_kernel(int *numOfElements, float *cellCenterX, float *cellCenterY, float *cellCenterZ, int *elementType,
		     int *neighborNum, size_t pitch,
		     float *speciesData, float *speciesData_F1, float *speciesData_F2, float *neighborData,
		     int numOfCells, int maxCells, int maxElements, float dt)
{
   int cellNum = blockIdx.x * blockDim.x + threadIdx.x;
   int j;

   if (cellNum >= numOfCells) return;

   // averages from neighbors
   float neighbor_NOTCH = 0.0;
   float neighbor_DELTA = 0.0;
   float neighbor_BOUND = 0.0;
   int numOfNeighbors = 0;

   for (j = 0; j < numOfCells; ++j) {
     if (j == cellNum) continue;

     if (distance_check(cellCenterX[cellNum], cellCenterY[cellNum], cellCenterZ[cellNum],
			cellCenterX[j], cellCenterY[j], cellCenterZ[j], S_XY_TOT, NEIGHBOR_DIST) != S_NONE) {
       ++numOfNeighbors;
       neighbor_NOTCH += speciesData[NOTCH_species*pitch+j];
       neighbor_DELTA += speciesData[DELTA_species*pitch+j];
       neighbor_BOUND += speciesData[BOUND_species*pitch+j];
     }
   }

   neighborNum[cellNum] = numOfNeighbors;
   neighborData[NOTCH_species*pitch+cellNum] = neighbor_NOTCH;
   neighborData[DELTA_species*pitch+cellNum] = neighbor_DELTA;
   neighborData[BOUND_species*pitch+cellNum] = neighbor_BOUND;
}


//
// Intracellular gene network
//

// F1 kernel for 2nd order Runge-Kutta
__global__ void
skin_kernel_F1(int *numOfElements, float *X, float *Y, float *Z, int *elementType,
	       int *neighborNum, size_t pitch,
	       float *speciesData, float *speciesData_F1, float *speciesData_F2, float *neighborData,
	       int numOfCells, int maxCells, int maxElements, float dt)
{
   int cellNum = blockIdx.x * blockDim.x + threadIdx.x;
   //int elemNum;

   if (cellNum >= numOfCells) return;

   float interactionValue, F1_val;

   // calculate F1
   float NOTCH_val = speciesData[NOTCH_species*pitch+cellNum];
   float DELTA_val = speciesData[DELTA_species*pitch+cellNum];
   float BOUND_val = speciesData[BOUND_species*pitch+cellNum];
   float BMA_val = speciesData[BMA_species*pitch+cellNum];
   float OVOL1_val = speciesData[OVOL1_species*pitch+cellNum];
   float OVOL2_val = speciesData[OVOL2_species*pitch+cellNum];
   float CMYC_val = speciesData[CMYC_species*pitch+cellNum];

   // averages from neighbors
   float neighbor_NOTCH = 0.0;
   float neighbor_DELTA = 0.0;
   float neighbor_BOUND = 0.0;
   int numOfNeighbors = 0;

   neighbor_NOTCH = neighborData[NOTCH_species*pitch+cellNum];
   neighbor_DELTA = neighborData[DELTA_species*pitch+cellNum];
   neighbor_BOUND = neighborData[BOUND_species*pitch+cellNum];
   numOfNeighbors = neighborNum[cellNum];
   if (numOfNeighbors != 0) {
     neighbor_NOTCH = neighbor_NOTCH / (float)numOfNeighbors;
     neighbor_DELTA = neighbor_DELTA / (float)numOfNeighbors;
     neighbor_BOUND = neighbor_BOUND / (float)numOfNeighbors;
   }

   //printf("%d %f %f %f\n", cellNum, neighbor_NOTCH, neighbor_DELTA, neighbor_BOUND);

   // NOTCH

   interactionValue = HILL(BOUND_val, NOTCH_pmin_BOUND, NOTCH_pmax_BOUND, NOTCH_c_BOUND, NOTCH_h_BOUND);
   interactionValue *= HILL(OVOL2_val, NOTCH_pmin_OVOL2, NOTCH_pmax_OVOL2, NOTCH_c_OVOL2, NOTCH_h_OVOL2);
   //interactionValue *= HILL(0.2f, NOTCH_pmin_OVOL2, NOTCH_pmax_OVOL2, NOTCH_c_OVOL2, NOTCH_h_OVOL2);
   F1_val = NOTCH_val + 0.5 * dt * ((-KA) * NOTCH_val * neighbor_DELTA + KD * BOUND_val - DF * NOTCH_val + interactionValue);
   if (F1_val < 0.0) F1_val = 0;
   speciesData_F1[NOTCH_species*pitch+cellNum] = F1_val;

   // DELTA

   //interactionValue = HILL(BOUND_val, DELTA_pmin_BOUND, DELTA_pmax_BOUND, DELTA_c_BOUND, DELTA_h_BOUND);
   interactionValue = HILL(NOTCH_val, DELTA_pmin_BOUND, DELTA_pmax_BOUND, DELTA_c_BOUND, DELTA_h_BOUND);
   F1_val = DELTA_val + 0.5 * dt * ((-KA) * DELTA_val * neighbor_NOTCH + KD * neighbor_BOUND - DA * DELTA_val + interactionValue);
   if (F1_val < 0.0) F1_val = 0;
   speciesData_F1[DELTA_species*pitch+cellNum] = F1_val;

   // BOUND

   F1_val = BOUND_val + 0.5 * dt * (KA * NOTCH_val * neighbor_DELTA - KD * BOUND_val - KI * BOUND_val);
   if (F1_val < 0.0) F1_val = 0;
   speciesData_F1[BOUND_species*pitch+cellNum] = F1_val;

   // BMA

   speciesData_F1[BMA_species*pitch+cellNum] = speciesData[BMA_species*pitch+cellNum];

   // OVOL1

   interactionValue = HILL(OVOL2_val, OVOL1_pmin_OVOL2, OVOL1_pmax_OVOL2, OVOL1_c_OVOL2, OVOL1_h_OVOL2);
   interactionValue *= HILL(BMA_val, OVOL1_pmin_BMA, OVOL1_pmax_BMA, OVOL1_c_BMA, OVOL1_h_BMA);
   F1_val = OVOL1_val + 0.5 * dt * (interactionValue - OVOL1_decay * OVOL1_val);
   if (F1_val < 0.0) F1_val = 0;
   speciesData_F1[OVOL1_species*pitch+cellNum] = F1_val;

   // OVOL2

   interactionValue = HILL(OVOL1_val, OVOL2_pmin_OVOL1, OVOL2_pmax_OVOL1, OVOL2_c_OVOL1, OVOL2_h_OVOL1);
   interactionValue *= HILL(0.4, 0.0, 1.0, 1.0, 1.0); // TGF-beta
   F1_val = OVOL2_val + 0.5 * dt * (interactionValue - OVOL2_decay * OVOL2_val);
   if (F1_val < 0.0) F1_val = 0;
   speciesData_F1[OVOL2_species*pitch+cellNum] = F1_val;

   // CMYC

   interactionValue = HILL(OVOL2_val, CMYC_pmin_OVOL2, CMYC_pmax_OVOL2, CMYC_c_OVOL2, CMYC_h_OVOL2);
   F1_val = CMYC_val + 0.5 * dt * (interactionValue - CMYC_decay * CMYC_val);
   if (F1_val < 0.0) F1_val = 0;
   speciesData_F1[CMYC_species*pitch+cellNum] = F1_val;
}

// F2 kernel for 2nd order Runge-Kutta
__global__ void
skin_kernel_F2(int *numOfElements, float *X, float *Y, float *Z, int *elementType,
	       int *neighborNum, size_t pitch,
	       float *speciesData, float *speciesData_F1, float *speciesData_F2, float *neighborData,
	       int numOfCells, int maxCells, int maxElements, float dt)
{
   int cellNum = blockIdx.x * blockDim.x + threadIdx.x;
   //int elemNum;

   if (cellNum >= numOfCells) return;

   float interactionValue, F2_val;

   // calculate F1
   float NOTCH_val = speciesData_F1[NOTCH_species*pitch+cellNum];
   float DELTA_val = speciesData_F1[DELTA_species*pitch+cellNum];
   float BOUND_val = speciesData_F1[BOUND_species*pitch+cellNum];
   float BMA_val = speciesData_F1[BMA_species*pitch+cellNum];
   float OVOL1_val = speciesData_F1[OVOL1_species*pitch+cellNum];
   float OVOL2_val = speciesData_F1[OVOL2_species*pitch+cellNum];
   float CMYC_val = speciesData_F1[CMYC_species*pitch+cellNum];

   // averages from neighbors
   float neighbor_NOTCH = 0.0;
   float neighbor_DELTA = 0.0;
   float neighbor_BOUND = 0.0;
   int numOfNeighbors = 0;

   neighbor_NOTCH = neighborData[NOTCH_species*pitch+cellNum];
   neighbor_DELTA = neighborData[DELTA_species*pitch+cellNum];
   neighbor_BOUND = neighborData[BOUND_species*pitch+cellNum];
   numOfNeighbors = neighborNum[cellNum];
   if (numOfNeighbors != 0) {
     neighbor_NOTCH = neighbor_NOTCH / (float)numOfNeighbors;
     neighbor_DELTA = neighbor_DELTA / (float)numOfNeighbors;
     neighbor_BOUND = neighbor_BOUND / (float)numOfNeighbors;
   }

   // NOTCH

   interactionValue = HILL(BOUND_val, NOTCH_pmin_BOUND, NOTCH_pmax_BOUND, NOTCH_c_BOUND, NOTCH_h_BOUND);
   interactionValue *= HILL(OVOL2_val, NOTCH_pmin_OVOL2, NOTCH_pmax_OVOL2, NOTCH_c_OVOL2, NOTCH_h_OVOL2);
   //interactionValue *= HILL(0.2f, NOTCH_pmin_OVOL2, NOTCH_pmax_OVOL2, NOTCH_c_OVOL2, NOTCH_h_OVOL2);
   F2_val = speciesData[NOTCH_species*pitch+cellNum] + dt * ((-KA) * NOTCH_val * neighbor_DELTA + KD * BOUND_val - DF * NOTCH_val + interactionValue);
   if (F2_val < 0.0) F2_val = 0;
   speciesData_F2[NOTCH_species*pitch+cellNum] = F2_val;

   // DELTA

   //interactionValue = HILL(BOUND_val, DELTA_pmin_BOUND, DELTA_pmax_BOUND, DELTA_c_BOUND, DELTA_h_BOUND);
   interactionValue = HILL(NOTCH_val, DELTA_pmin_BOUND, DELTA_pmax_BOUND, DELTA_c_BOUND, DELTA_h_BOUND);
   F2_val = speciesData[DELTA_species*pitch+cellNum] + dt * ((-KA) * DELTA_val * neighbor_NOTCH + KD * neighbor_BOUND - DA * DELTA_val + interactionValue);
   if (F2_val < 0.0) F2_val = 0;
   speciesData_F2[DELTA_species*pitch+cellNum] = F2_val;

   // BOUND

   F2_val = speciesData[BOUND_species*pitch+cellNum] + dt * (KA * NOTCH_val * neighbor_DELTA - KD * BOUND_val - KI * BOUND_val);
   if (F2_val < 0.0) F2_val = 0;
   speciesData_F2[BOUND_species*pitch+cellNum] = F2_val;

   // BMA

   speciesData_F2[BMA_species*pitch+cellNum] = speciesData_F1[BMA_species*pitch+cellNum];

   // OVOL1

   interactionValue = HILL(OVOL2_val, OVOL1_pmin_OVOL2, OVOL1_pmax_OVOL2, OVOL1_c_OVOL2, OVOL1_h_OVOL2);
   interactionValue *= HILL(BMA_val, OVOL1_pmin_BMA, OVOL1_pmax_BMA, OVOL1_c_BMA, OVOL1_h_BMA);
   F2_val = speciesData[OVOL1_species*pitch+cellNum] + dt * (interactionValue - OVOL1_decay * OVOL1_val);
   if (F2_val < 0.0) F2_val = 0;
   speciesData_F2[OVOL1_species*pitch+cellNum] = F2_val;

   // OVOL2

   interactionValue = HILL(OVOL1_val, OVOL2_pmin_OVOL1, OVOL2_pmax_OVOL1, OVOL2_c_OVOL1, OVOL2_h_OVOL1);
   interactionValue *= HILL(0.4, 0.0, 1.0, 1.0, 1.0); // TGF-beta
   F2_val = speciesData[OVOL2_species*pitch+cellNum] + dt * (interactionValue - OVOL2_decay * OVOL2_val);
   if (F2_val < 0.0) F2_val = 0;
   speciesData_F2[OVOL2_species*pitch+cellNum] = F2_val;

   // CMYC

   interactionValue = HILL(OVOL2_val, CMYC_pmin_OVOL2, CMYC_pmax_OVOL2, CMYC_c_OVOL2, CMYC_h_OVOL2);
   F2_val = speciesData[CMYC_species*pitch+cellNum] + dt * (interactionValue - CMYC_decay * CMYC_val);
   if (F2_val < 0.0) F2_val = 0;
   speciesData_F2[CMYC_species*pitch+cellNum] = F2_val;
}
