/*
 sem.cu
 
 Main functions for running the GPU kernels.

 Author: Scott Christley <schristley@mac.com>
 
 Copyright (C) 2010 Scott Christley
 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met: 1. Redistributions of source code must retain the above
 copyright notice, this list of conditions and the following
 disclaimer.  2. Redistributions in binary form must reproduce the
 above copyright notice, this list of conditions and the following
 disclaimer in the documentation and/or other materials provided with
 the distribution.  3. The name of the author may not be used to
 endorse or promote products derived from this software without
 specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 */

#include <stdio.h>

// Constant GPU array for holding parameters
__constant__ float skin_parameters[100];

extern "C" {
#include "GPUDefines.h"
#include "sem_kernel.cu"

// structure to hold all of the different data together
typedef struct _sem_GPUgrids {
  int maxCells;
  int numOfCells;
  int maxElements;
  float dt;
  int *numOfElements;
  int *elementType;
  float *X;
  float *X_F1;
  float *X_F2;
  float *Y;
  float *Y_F1;
  float *Y_F2;
  float *Z;
  float *Z_F1;
  float *Z_F2;
  size_t pitch;
  float *cellCenterX;
  float *cellCenterY;
  float *cellCenterZ;
} sem_GPUgrids;


// Allocation
void *sem_allocGPUKernel(void *model, int maxCells, int maxElements, float dt, float *hostParameters)
{
 sem_GPUgrids *grids = (sem_GPUgrids *)malloc(sizeof(sem_GPUgrids));

 // Save parameters
 grids->maxCells = maxCells;
 grids->maxElements = maxElements;
 grids->dt = dt;

 // Allocate device memory

 // cells and elements
 cudaMalloc((void**)&(grids->numOfElements), maxCells * sizeof(int));
 cudaMallocPitch((void**)&(grids->elementType), &(grids->pitch), maxCells * sizeof(int), maxElements);
 cudaMallocPitch((void**)&(grids->X), &(grids->pitch), maxCells * sizeof(float), maxElements);
 cudaMallocPitch((void**)&(grids->X_F1), &(grids->pitch), maxCells * sizeof(float), maxElements);
 cudaMallocPitch((void**)&(grids->X_F2), &(grids->pitch), maxCells * sizeof(float), maxElements);
 cudaMallocPitch((void**)&(grids->Y), &(grids->pitch), maxCells * sizeof(float), maxElements);
 cudaMallocPitch((void**)&(grids->Y_F1), &(grids->pitch), maxCells * sizeof(float), maxElements);
 cudaMallocPitch((void**)&(grids->Y_F2), &(grids->pitch), maxCells * sizeof(float), maxElements);
 cudaMallocPitch((void**)&(grids->Z), &(grids->pitch), maxCells * sizeof(float), maxElements);
 cudaMallocPitch((void**)&(grids->Z_F1), &(grids->pitch), maxCells * sizeof(float), maxElements);
 cudaMallocPitch((void**)&(grids->Z_F2), &(grids->pitch), maxCells * sizeof(float), maxElements);

 // cell centers
 cudaMalloc((void**)&(grids->cellCenterX), maxCells * sizeof(float));
 cudaMalloc((void**)&(grids->cellCenterY), maxCells * sizeof(float));
 cudaMalloc((void**)&(grids->cellCenterZ), maxCells * sizeof(float));

 // copy parameters
 cudaMemcpyToSymbol(skin_parameters, hostParameters, 100 * sizeof(float), 0, cudaMemcpyHostToDevice);

 return grids;
}

// Initialization
void sem_initGPUKernel(void *model, void *g, int numOfCells, int *numOfElements, void *hostX, void *hostY, void *hostZ, void *hostType)
{
 sem_GPUgrids *grids = (sem_GPUgrids *)g;

 grids->numOfCells = numOfCells;

 // Copy host memory to device memory
 cudaMemcpy(grids->numOfElements, numOfElements, grids->maxCells * sizeof(int), cudaMemcpyHostToDevice);
 cudaMemcpy2D(grids->elementType, grids->pitch, hostType, grids->maxCells * sizeof(int), grids->maxCells * sizeof(int), grids->maxElements, cudaMemcpyHostToDevice);
 cudaMemcpy2D(grids->X, grids->pitch, hostX, grids->maxCells * sizeof(float), grids->maxCells * sizeof(float), grids->maxElements, cudaMemcpyHostToDevice);
 cudaMemcpy2D(grids->Y, grids->pitch, hostY, grids->maxCells * sizeof(float), grids->maxCells * sizeof(float), grids->maxElements, cudaMemcpyHostToDevice);
 cudaMemcpy2D(grids->Z, grids->pitch, hostZ, grids->maxCells * sizeof(float), grids->maxCells * sizeof(float), grids->maxElements, cudaMemcpyHostToDevice);
}

// Execution, just cell movement
void sem_invokeGPUKernel(void *model, void *g, void *hostX, void *hostY, void *hostZ, int timeSteps)
{
 int t;
 sem_GPUgrids *grids = (sem_GPUgrids *)g;

 // Invoke kernel
 int x = 64;
 int y = 4;

 int threadsPerBlock1D = x;
 int blocksPerGrid1D = (grids->numOfCells + threadsPerBlock1D - 1) / threadsPerBlock1D;

 dim3 threadsPerBlock(x, y);
 dim3 threadsPerGrid((grids->numOfCells + threadsPerBlock.x - 1) / threadsPerBlock.x, (grids->maxElements + threadsPerBlock.y - 1) / threadsPerBlock.y);

 for (t = 0; t < timeSteps; ++t) {

#if 1
   // this is the faster cell center implementation
  skin_center_kernel<<< blocksPerGrid1D, threadsPerBlock1D >>>(grids->numOfElements, grids->X, grids->Y, grids->Z, grids->elementType,
							       grids->cellCenterX, grids->cellCenterY, grids->cellCenterZ, grids->pitch/sizeof(float),
							       grids->numOfCells, grids->maxCells, grids->maxElements, grids->dt);

  skin_moveKernel_F1<<< threadsPerGrid, threadsPerBlock >>>(grids->numOfElements, grids->X, grids->X_F1, grids->X_F2, grids->Y, grids->Y_F1, grids->Y_F2,
							    grids->Z, grids->Z_F1, grids->Z_F2, grids->cellCenterX, grids->cellCenterY, grids->cellCenterZ, grids->elementType,
							    grids->pitch/sizeof(float), grids->numOfCells, grids->maxCells, grids->maxElements, grids->dt);

  skin_moveKernel_F2<<< threadsPerGrid, threadsPerBlock >>>(grids->numOfElements, grids->X, grids->X_F1, grids->X_F2, grids->Y, grids->Y_F1, grids->Y_F2,
							    grids->Z, grids->Z_F1, grids->Z_F2, grids->cellCenterX, grids->cellCenterY, grids->cellCenterZ, grids->elementType,
							    grids->pitch/sizeof(float), grids->numOfCells, grids->maxCells, grids->maxElements, grids->dt);
#else
  // the slower naive implementation
   sem_kernel_F1<<< threadsPerGrid, threadsPerBlock >>>(grids->numOfElements, grids->X, grids->X_F1, grids->X_F2, grids->Y, grids->Y_F1, grids->Y_F2,
							grids->Z, grids->Z_F1, grids->Z_F2, grids->elementType,
							grids->pitch/sizeof(float), grids->numOfCells, grids->maxCells, grids->maxElements, grids->dt);

   sem_kernel_F2<<< threadsPerGrid, threadsPerBlock >>>(grids->numOfElements, grids->X, grids->X_F1, grids->X_F2, grids->Y, grids->Y_F1, grids->Y_F2,
							grids->Z, grids->Z_F1, grids->Z_F2, grids->elementType,
							grids->pitch/sizeof(float), grids->numOfCells, grids->maxCells, grids->maxElements, grids->dt);
#endif

   cudaMemcpy2D(grids->X, grids->pitch, grids->X_F2, grids->pitch, grids->maxCells * sizeof(float), grids->maxElements, cudaMemcpyDeviceToDevice);
   cudaMemcpy2D(grids->Y, grids->pitch, grids->Y_F2, grids->pitch, grids->maxCells * sizeof(float), grids->maxElements, cudaMemcpyDeviceToDevice);
   cudaMemcpy2D(grids->Z, grids->pitch, grids->Z_F2, grids->pitch, grids->maxCells * sizeof(float), grids->maxElements, cudaMemcpyDeviceToDevice);
 }

 // Copy result to host memory
 cudaMemcpy2D(hostX, grids->maxCells * sizeof(float), grids->X, grids->pitch, grids->maxCells * sizeof(float), grids->maxElements, cudaMemcpyDeviceToHost);
 cudaMemcpy2D(hostY, grids->maxCells * sizeof(float), grids->Y, grids->pitch, grids->maxCells * sizeof(float), grids->maxElements, cudaMemcpyDeviceToHost);
 cudaMemcpy2D(hostZ, grids->maxCells * sizeof(float), grids->Z, grids->pitch, grids->maxCells * sizeof(float), grids->maxElements, cudaMemcpyDeviceToHost);
}


// Release
void sem_releaseGPUKernel(void *model, void *g)
{
 sem_GPUgrids *grids = (sem_GPUgrids *)g;
 cudaFree(grids->numOfElements);
 cudaFree(grids->elementType);
 cudaFree(grids->X);
 cudaFree(grids->X_F1);
 cudaFree(grids->X_F2);
 cudaFree(grids->Y);
 cudaFree(grids->Y_F1);
 cudaFree(grids->Y_F2);
 cudaFree(grids->Z);
 cudaFree(grids->Z_F1);
 cudaFree(grids->Z_F2);
 cudaFree(grids->cellCenterX);
 cudaFree(grids->cellCenterY);
 cudaFree(grids->cellCenterZ);
 free(grids);
 cudaThreadExit();
}

}
